# Copyright (c) 2021-2024 SuperCowPowers LLC

"""
Workbench Bridges
- TBD
  - TBD1
  - TBD2
"""
from importlib.metadata import version

try:
    __version__ = version("workbench_bridges")
except Exception:
    __version__ = "unknown"

# Workbench Bridges Logging
from workbench_bridges.utils.logger import logging_setup

logging_setup()



"""Welcome to the Workbench-Bridges API Classes

- ParameterStore: Manages AWS Parameter Store
- InferenceStore: Manages Athena based storage for inference results
- DFStore: Manages DataFrames in AWS S3
"""

from .parameter_store import ParameterStore
from .inference_store import InferenceStore
from .df_store import DFStore

__all__ = [
    "ParameterStore",
    "InferenceStore",
    "DFStore",
]



"""Fast Inference on SageMaker Endpoints"""

import os
import pandas as pd
from io import StringIO
import logging
from concurrent.futures import ThreadPoolExecutor

# Sagemaker Imports
import boto3
import sagemaker
from sagemaker.serializers import CSVSerializer
from sagemaker.deserializers import CSVDeserializer
from sagemaker import Predictor

log = logging.getLogger("workbench-bridges")

_CACHED_SM_SESSION = None


def get_aws_region():
    # Try environment variables first
    region = os.environ.get("SAGEMAKER_REGION") or os.environ.get("AWS_REGION") or os.environ.get("AWS_DEFAULT_REGION")
    print(f"ENV REGION: {region}")
    region = region or boto3.session.Session().region_name
    if not region:
        msg = "No AWS region configured! Please set SAGEMAKER_REGION or AWS_REGION"
        log.critical(msg)
        raise Exception(msg)
    return region


def get_or_create_sm_session():
    global _CACHED_SM_SESSION
    if _CACHED_SM_SESSION is None:
        region = get_aws_region()
        print(f"Creating new SageMaker session in region: {region}")
        _CACHED_SM_SESSION = sagemaker.Session(boto3.Session(region_name=region))
    return _CACHED_SM_SESSION


def fast_inference(endpoint_name: str, eval_df: pd.DataFrame, sm_session=None, threads: int = 4) -> pd.DataFrame:
    """Run inference on the Endpoint using the provided DataFrame

    Args:
        endpoint_name (str): The name of the Endpoint
        eval_df (pd.DataFrame): The DataFrame to run predictions on
        sm_session (sagemaker.session.Session, optional): SageMaker Session. If None, a cached session is created.
        threads (int): The number of threads to use (default: 4)

    Returns:
        pd.DataFrame: The DataFrame with predictions
    """
    # Use cached session if none is provided
    if sm_session is None:
        sm_session = get_or_create_sm_session()

    predictor = Predictor(
        endpoint_name,
        sagemaker_session=sm_session,
        serializer=CSVSerializer(),
        deserializer=CSVDeserializer(),
    )

    total_rows = len(eval_df)

    # Fixed chunk size of 100 rows
    chunk_size = 100

    def process_chunk(chunk_df: pd.DataFrame, start_index: int) -> pd.DataFrame:
        log.info(f"Processing {start_index}:{min(start_index + chunk_size, total_rows)} out of {total_rows} rows...")
        csv_buffer = StringIO()
        chunk_df.to_csv(csv_buffer, index=False)
        try:
            response = predictor.predict(csv_buffer.getvalue())
        except Exception as e:
            log.error(f"Error during prediction: {e}")
            return pd.DataFrame()
        # CSVDeserializer returns a nested list: first row is headers
        return pd.DataFrame.from_records(response[1:], columns=response[0])

    # Sagemaker has a connection pool limit of 10
    if threads > 10:
        log.warning("Sagemaker has a connection pool limit of 10. Reducing threads to 10.")
        threads = 10

    # Split DataFrame into chunks and process them concurrently
    chunks = [(eval_df[i : i + chunk_size], i) for i in range(0, total_rows, chunk_size)]  # noqa: E203

    # Use min of threads or number of chunks to avoid creating unnecessary threads
    actual_threads = min(threads, len(chunks))
    with ThreadPoolExecutor(max_workers=actual_threads) as executor:
        df_list = list(executor.map(lambda p: process_chunk(*p), chunks))

    # Filter out empty DataFrames that might result from errors
    df_list = [df for df in df_list if not df.empty]

    if not df_list:
        log.error("All prediction chunks failed")
        return pd.DataFrame()

    combined_df = pd.concat(df_list, ignore_index=True)

    # Convert the column types of the dataframe (CSVDeserializer returns all columns as strings/objects)
    combined_df = df_type_conversions(combined_df)
    return combined_df


def df_type_conversions(df: pd.DataFrame) -> pd.DataFrame:
    """Convert the types of the dataframe that we get from an endpoint

    Args:
        df (pd.DataFrame): DataFrame to convert

    Returns:
        pd.DataFrame: Converted DataFrame
    """
    # Some endpoints will put in "N/A" values (for CSV serialization)
    # We need to convert these to NaN and the run the conversions below
    # Report on the number of N/A values in each column in the DataFrame
    # For any count above 0 list the column name and the number of N/A values
    na_counts = df.isin(["N/A"]).sum()
    for column, count in na_counts.items():
        if count > 0:
            log.warning(f"{column} has {count} N/A values, converting to NaN")
    pd.set_option("future.no_silent_downcasting", True)
    df = df.replace("N/A", float("nan"))

    # Convert data to numeric
    # Note: Since we're using CSV serializers numeric columns often get changed to generic 'object' types

    # Hard Conversion
    # Note: We explicitly catch exceptions for columns that cannot be converted to numeric
    for column in df.columns:
        try:
            df[column] = pd.to_numeric(df[column])
        except ValueError:
            # If a ValueError is raised, the column cannot be converted to numeric, so we keep it as is
            pass
        except TypeError:
            # This typically means a duplicated column name, so confirm duplicate (more than 1) and log it
            column_count = (df.columns == column).sum()
            log.critical(f"{column} occurs {column_count} times in the DataFrame.")
            pass

    # Soft Conversion
    # Convert columns to the best possible dtype that supports the pd.NA missing value.
    df = df.convert_dtypes()

    # Convert pd.NA placeholders to pd.NA
    # Note: CSV serialization converts pd.NA to blank strings, so we have to put in placeholders
    df.replace("__NA__", pd.NA, inplace=True)

    # Check for True/False values in the string columns
    for column in df.select_dtypes(include=["string"]).columns:
        if df[column].str.lower().isin(["true", "false"]).all():
            df[column] = df[column].str.lower().map({"true": True, "false": False})

    # Return the Dataframe
    return df


if __name__ == "__main__":
    """Exercise the Endpoint Utilities"""
    import time
    from workbench.api.endpoint import Endpoint
    from workbench.utils.endpoint_utils import fs_evaluation_data

    # Grab the Endpoint
    my_endpoint_name = "abalone-regression"
    my_endpoint = Endpoint(my_endpoint_name)
    if not my_endpoint.exists():
        print(f"Endpoint {my_endpoint_name} does not exist.")
        exit(1)

    # Pull evaluation data
    print("Pulling Evaluation Data...")
    sagemaker_session = my_endpoint.sm_session
    my_eval_df = fs_evaluation_data(my_endpoint)
    start_time = time.time()
    my_results_df = fast_inference(my_endpoint_name, my_eval_df, sagemaker_session)
    end_time = time.time()
    print(f"Fast Inference took {end_time - start_time} seconds")
    print(my_results_df)
    print(my_results_df.info())

    # Test with no session
    my_results_df = fast_inference(my_endpoint_name, my_eval_df)
    print(my_results_df)
    print(my_results_df.info())



import os
import sys
import time
import pandas as pd
import logging
import argparse

# SageMaker-Bridges Imports
from workbench_bridges.endpoints.fast_inference import fast_inference

# Set up logging
log = logging.getLogger()


def download_data(endpoint_name: str):
    """Download the data Workbench FeatureSet

    Args:
        endpoint_name (str): The name of the Endpoint
    """
    from workbench.api import FeatureSet, Model, Endpoint

    fs = FeatureSet(Model(Endpoint(endpoint_name).get_input()).get_input())
    df = fs.pull_dataframe()
    df.to_csv("test_evaluation_data.csv", index=False)


def launch_inference(data: pd.DataFrame, endpoint_name: str):
    """Launch inference on the provided data.

    Args:
        data (pd.DataFrame): Data for inference
        endpoint_name (str): The name of the endpoint
    """
    num_rows = 1000

    # Sample data for inference
    data_sample = data.sample(n=num_rows, replace=True)
    print(f"\nTiming Inference on {len(data_sample)} rows")

    start_time = time.time()
    results = fast_inference(endpoint_name, data_sample)
    inference_time = time.time() - start_time

    print(f"Inference Time: {inference_time} on Endpoint: {endpoint_name}")

    # Print out the results
    print("\nInference Results:")
    print(results)


if __name__ == "__main__":
    # Set up argument parser
    parser = argparse.ArgumentParser(description="Launch an AWS Inference Run")
    args = parser.parse_args()

    # Endpoint name to test
    test_endpoint_name = "test-timing-realtime"

    # Track total time
    total_start_time = time.time()

    # Check if we have local data
    if not os.path.exists("test_evaluation_data.csv"):
        log.warning("Downloading Data... Rerun the script after the download completes")
        download_data(test_endpoint_name)
        sys.exit(1)

    # Local data this will duplicate a launch from an App like LiveDesign/StarDrop
    data = pd.read_csv("test_evaluation_data.csv")

    # Launch inference
    launch_inference(data, test_endpoint_name)

    # Track total time
    total_time = time.time() - total_start_time
    print(f"Total Time: {total_time}")



import os
import sys
import pandas as pd

# Workbench-Bridges Imports
from workbench_bridges.endpoints.fast_inference import fast_inference


if __name__ == "__main__":

    # Check if we have local data
    if not os.path.exists("test_evaluation_data.csv"):
        print("Download Data to test_evaluation_data.csv")
        sys.exit(1)

    # Data will be passed in from the End-User Application
    eval_df = pd.read_csv("test_evaluation_data.csv")[:1000]

    # Run inference on AWS Endpoint
    endpoint_name = "test-timing-realtime"
    results = fast_inference(endpoint_name, eval_df)

    # A Dataframe with Predictions is returned
    print(results)



"""Helper functions for working with ISO-8601 formatted dates and times"""

from datetime import datetime, date, timezone
import numpy as np
import logging
import time

log = logging.getLogger("workbench")

# A simple log throttle for specific log messages
last_log = 0
log_interval = 5  # seconds


def datetime_to_iso8601(datetime_obj) -> str:
    """Convert datetime or date to string in UTC format (yyyy-MM-dd'T'HH:mm:ss.SSSZ)
    Args:
        datetime_obj (datetime/date): The datetime or date object to convert
    Returns:
        str: The datetime as a string in ISO-8601 format
    Note: This particular format is required by AWS Feature Store
    """
    global last_log
    current_time = time.time()

    # Check for valid input
    if not isinstance(datetime_obj, (datetime, date)):
        log.error("Invalid input: Expected datetime or date object")
        return None

    # Convert date to datetime if needed
    if isinstance(datetime_obj, date) and not isinstance(datetime_obj, datetime):
        log.info("Received date object; converting to datetime at midnight UTC.")
        datetime_obj = datetime.combine(datetime_obj, datetime.min.time()).replace(tzinfo=timezone.utc)

    # Check for TimeZone
    if datetime_obj.tzinfo is None:
        if current_time - last_log >= log_interval:
            log.warning("Datetime object is naive; localizing to UTC.")
            last_log = current_time
        datetime_obj = datetime_obj.replace(tzinfo=timezone.utc)

    try:
        # Convert to ISO-8601 String
        iso_str = datetime_obj.astimezone(timezone.utc).isoformat("T", "milliseconds")
        return iso_str.replace("+00:00", "Z")
    except Exception as e:
        log.error(f"Failed to convert datetime to ISO-8601 string: {e}")
        return None


def iso8601_to_datetime(iso8601_str: str) -> datetime:
    """Convert ISO-8601 string to datetime object
    Args:
        iso8601_str (str): The ISO-8601 string to convert
    Returns:
        datetime: The datetime object
    """
    if "Z" in iso8601_str:
        iso8601_str = iso8601_str.replace("Z", "+00:00")
    return datetime.fromisoformat(iso8601_str).replace(tzinfo=timezone.utc)


def convert_all_to_iso8601(data):
    """Convert datetime fields to ISO-8601 string
    Args:
        data (arbitrary type): The data to convert
    Returns:
        arbitrary type: The converted data
    """
    if isinstance(data, dict):
        result = {}
        for key, value in data.items():
            result[key] = convert_all_to_iso8601(value)
        return result
    elif isinstance(data, list):
        result = []
        for item in data:
            result.append(convert_all_to_iso8601(item))
        return result
    elif isinstance(data, datetime):
        return datetime_to_iso8601(data)
    elif isinstance(data, np.int64):
        return int(data)
    else:
        return data


def datetime_string(datetime_obj: datetime) -> str:
    """Helper: Convert DateTime Object into a nicely formatted string.

    Args:
        datetime_obj (datetime): The datetime object to convert.

    Returns:
        str: The datetime as a string in the format "YYYY-MM-DD HH:MM", or "-" if input is None or "-".
    """
    placeholder = "-"
    if datetime_obj is None or datetime_obj == placeholder:
        return placeholder

    if not isinstance(datetime_obj, datetime):
        log.debug("Expected datetime object.. trying to convert...")
        try:
            datetime_obj = iso8601_to_datetime(datetime_obj)
        except Exception as e:
            log.error(f"Failed to convert datetime object: {e}")
            return str(datetime_obj)

    try:
        return datetime_obj.astimezone().strftime("%Y-%m-%d %H:%M")
    except Exception as e:
        log.error(f"Failed to convert datetime to string: {e}")
        return str(datetime_obj)


if __name__ == "__main__":
    """Exercise the helper functions"""

    # Test the conversion to ISO-8601
    now = datetime.now(timezone.utc)
    print(now)
    now_str = datetime_to_iso8601(now)
    print(now_str)

    # Test the conversion from ISO-8601 back to datetime
    now2 = iso8601_to_datetime(now_str)
    print(now2)

    # Test the conversion of all datetime fields to ISO-8601
    data = {"a": 1, "b": "2", "c": now, "d": {"e": 3, "f": now}}
    print(data)
    data2 = convert_all_to_iso8601(data)
    print(data2)

    # Test the datetime string conversion
    print(datetime_string(now))



"""ExecutionEnvironment provides logic/functionality to figure out the current execution environment"""

import os
import sys
import logging
import requests

# Workbench Bridge Imports
from workbench_bridges.utils.glue_utils import get_resolved_options

# Set up the logger
log = logging.getLogger("workbench-bridges")


def running_on_glue() -> bool:
    """
    Check if the current execution environment is an AWS Glue job.

    Returns:
        bool: True if running in AWS Glue environment, False otherwise.
    """
    return bool(os.environ.get("GLUE_VERSION") or os.environ.get("GLUE_PYTHON_VERSION"))


def running_on_lambda() -> bool:
    """
    Check if the current execution environment is an AWS Lambda function.

    Returns:
        bool: True if running in AWS Lambda environment, False otherwise.
    """
    return bool(os.environ.get("AWS_LAMBDA_FUNCTION_NAME"))


def running_on_ecs() -> bool:
    """
    Check if the current environment is running on AWS ECS.

    Returns:
        bool: True if running on AWS ECS, False otherwise.
    """
    indicators = [
        "ECS_SERVICE_NAME",
        "ECS_CONTAINER_METADATA_URI",
        "ECS_CONTAINER_METADATA_URI_V4",
        "AWS_CONTAINER_CREDENTIALS_RELATIVE_URI",
        "AWS_EXECUTION_ENV",
    ]
    return any(os.environ.get(indicator) for indicator in indicators)


def running_on_docker() -> bool:
    """Check if the current environment is running on a Docker container.

    Returns:
        bool: True if running in a Docker container, False otherwise.
    """
    # Check for .dockerenv file
    if os.path.exists("/.dockerenv"):
        return True

    # Check cgroup for docker
    try:
        with open("/proc/self/cgroup") as f:
            if any("docker" in line for line in f):
                return True
    except (FileNotFoundError, PermissionError):
        pass

    return running_on_ecs()


def running_as_service() -> bool:
    """
    Check if the current environment is running as a service (e.g. Docker, ECS, Glue, Lambda).

    Returns:
        bool: True if running as a service, False otherwise.
    """
    return any([running_on_glue(), running_on_lambda(), running_on_ecs(), running_on_docker()])


def glue_job_name() -> str:
    """Get Glue job name from environment or script."""
    args = get_resolved_options(sys.argv)

    if job_name := args.get("JOB_NAME"):
        return job_name

    # Fallback to script name
    if script_location := args.get("scriptLocation"):
        return os.path.splitext(os.path.basename(script_location))[0]

    return "unknown"


def ecs_job_name() -> str:
    """Get ECS job name from metadata or environment."""
    # Try metadata endpoint first
    if metadata_uri := os.environ.get("ECS_CONTAINER_METADATA_URI_V4"):
        try:
            response = requests.get(f"{metadata_uri}/task", timeout=5)
            if response.status_code == 200:
                if family := response.json().get("Family"):
                    return family
        except requests.RequestException as e:
            log.warning(f"Failed to fetch ECS metadata: {e}")

    return os.environ.get("ECS_SERVICE_NAME", "unknown")


if __name__ == "__main__":
    """Test the Execution Environment utilities"""

    # Test running_on_glue
    assert running_on_glue() is False
    os.environ["GLUE_VERSION"] = "1.0"
    assert running_on_glue() is True
    del os.environ["GLUE_VERSION"]

    # Test running_on_lambda
    assert running_on_lambda() is False
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "my_lambda_function"
    assert running_on_lambda() is True
    del os.environ["AWS_LAMBDA_FUNCTION_NAME"]

    # Test running_on_docker
    assert running_on_docker() is False
    os.environ["ECS_CONTAINER_METADATA_URI"] = "http://localhost:8080"
    assert running_on_docker() is True
    del os.environ["ECS_CONTAINER_METADATA_URI"]

    # Test running_on_ecs
    assert running_on_ecs() is False
    os.environ["ECS_CONTAINER_METADATA_URI"] = "http://localhost:8080"
    assert running_on_ecs() is True
    del os.environ["ECS_CONTAINER_METADATA_URI"]

    # Test getting the Glue Job Name
    print(glue_job_name())

    print("All tests passed!")
